!=========================================================================
! Copyright (C) GemTalk Systems 1986-2017.  All Rights Reserved.
!
! $Id: ZeroDivide.gs 42264 2017-07-28 05:24:38Z lalmarod $
!
!=========================================================================

set class ZeroDivide

! no removeall 
! ZeroDivide>>initialize is in Exception.gs

category: 'For Documentation Installation only'
classmethod: 
installDocumentation

self comment:
'The class protocol describe the behavior of the global ZeroDivide. 
It is used to as an exception selector to catch zero divide exceptions 
and can also be used to signal that a division by zero error has 
occured. Zero divide exceptions are resumable so any message in this 
protocol that signal such an exception may ultimately return to their 
sender. The signaled exceptions generated by this type of object 
conform to the protocol <ZeroDivide>.

The instance protocol describes the behavior of exceptions that are 
signalled when an attempt is made to divide some number (the dividend) 
by zero.

Constraints:
	gsResumable: Boolean
	gsTrappable: Object
	gsNumber: SmallInteger
	currGsHandler: GsExceptionHandler
	gsStack: Object
	gsReason: String
	gsDetails: Object
	tag: Object
	messageText: Object
	gsArgs: Object
	dividend: Number'.
%

! -----------------------------------------
category: 'Instance creation'
classmethod: 
dividend: argument
	"Signal the occurrence of a division by zero exception. Capture the number
	that was being divided such that it is available from the 'signaled exception.'

	If the message #dividend is subsequently sent to the <ZeroDivide> object
	that is the 'signaled exception' the value of 'argument' is returned."

	"It seems rare that this Exception would be signaled by sending this message.
	Typically, this Exception is signaled indirectly by the VM when a 
	division by zero occurs. So why not just take the usual path?!"

	argument / 0.
%

classmethod: 
signal

	0 / 0.
%

! -----------------------------------------
category: 'Accessing'
method: 
dividend

	^ dividend
%

method: 
dividend: aNumber

  dividend := aNumber
%

! fixed 41582
method: 
asString
^ self _description: 
   'attempt to divide ' , dividend asString , ' by zero' 
%
